from typing import Optional, List, Tuple

import cv2
import numpy as np
from joblib import Parallel, delayed
from skimage import feature

from mmir import Point, Image, Match
from mmir.feature_detection import InterestPoints
from mmir.feature_detection.filters import DescriptorsFilter
from mmir.feature_matching import FeatureMatcher, Correlations
from mmir.math import window_in_bounds, roi, ssd


class HogFeatureMatcher(FeatureMatcher):
    """
    Uses HOG (Histogram of Oriented Gradients) to find matches between
    two sets of keypoints of two images.
    """

    def __init__(
            self,
            window_size: int = 40,
            filters: Optional[List[DescriptorsFilter]] = None,
            best_match_threshold: float = .9,
    ) -> None:
        """
        :param window_size:
            The size of the window centered in the keypoint used for computing the HOG
            which is used for feature matching.
        :param measure:
            The measure to be used for comparing the distance between two HOG feature
            vectors.
        :param filters:
            Filters to be used for deciding on the candidate descriptors for a match.
        """
        super().__init__(filters=filters)

        self._window_size = window_size
        self._filters = filters if filters else []
        self._best_match_threshold = best_match_threshold

    def find_match(
            self,
            matched_point: Point,
            left: Image,
            right: Image
    ) -> Optional[Tuple[Point, float]]:
        if not window_in_bounds(center=matched_point, window_size=self._window_size, shape=left.shape):
            return None

        keypoint_subregion = roi(
            center=matched_point,
            width=self._window_size,
            height=self._window_size,
            image=left
        )

        candidate_points = (
            [x, matched_point[1]] for x in range(0, right.shape[1])
        )

        candidate_points = (
            descriptor for descriptor in candidate_points if window_in_bounds(
                center=descriptor,
                window_size=self._window_size,
                shape=left.shape
            )
        )

        for filter in self._filters:
            candidate_points = filter(matched_point, candidate_points)

        reference_hog = feature.hog(keypoint_subregion, block_norm='L1')
        best_match = None
        for candidate_point in candidate_points:
            candidate_subregion = roi(
                center=candidate_point,
                width=self._window_size,
                height=self._window_size,
                image=right,
            )
            candidate_hog = feature.hog(candidate_subregion, block_norm='L1')
            distance = ssd(reference_hog, candidate_hog)
            if best_match is None or distance < best_match[1]:
                best_match = (candidate_point, distance)

        if self._best_match_threshold is None:
            return best_match

        if not best_match:
            return best_match

        if best_match[1] < self._best_match_threshold:
            return None

        return best_match

    def find_match_for(self, matched_point: Point, detector_output: InterestPoints) -> Optional[Match]:
        match = self.find_match(
            matched_point,
            detector_output.left,
            detector_output.right
        )

        if not match:
            return None

        return matched_point, match[0], match[1]

    @staticmethod
    def visualize(output: Correlations) -> Image:
        left = cv2.cvtColor(np.copy(output.left), cv2.COLOR_GRAY2BGR)
        right = cv2.cvtColor(np.copy(output.right), cv2.COLOR_GRAY2BGR)

        right = cv2.resize(src=right, dsize=(left.shape[1], left.shape[0]))
        img = np.concatenate((left, right), axis=1)
        for match in output.matches:
            pt1 = (int(match[0][0]), int(match[0][1]))
            pt2 = (int(match[1][0] + output.left.shape[1]), int(match[1][1]))

            cv2.line(img=img, pt1=pt1, pt2=pt2, color=(0, 255, 0))

        cv2.imshow("img", img)
        return output

    def __call__(self, detector_output: InterestPoints) -> Correlations:
        matches = Parallel(n_jobs=8)(
            delayed(self.find_match_for)(keypoint, detector_output) for keypoint in detector_output.keypoints_left
        )

        matches = list(filter(lambda x: x is not None, matches))
        matches = sorted(matches, key=lambda x: x[2], reverse=True)

        return Correlations(detector_output.left, detector_output.right, matches)



from typing import List, Generator

from mmir.math.epipolar import get_epipolar_line
from mmir.math.math import point_to_line_distance
from mmir import Point, Matrix
from .descriptor_filter import DescriptorsFilter


class EpipolarDistanceFilter(DescriptorsFilter):
    """
    Filters out all the candidate points which are further than a given distance
    away from the epipolar line of the point that is being matched.
    """
    def __init__(self, fundamental_matrix: Matrix, max_distance=5, *args, **kwargs):
        """
        :param fundamental_matrix:
            The fundamental matrix to be used when computing the epipolar line of the point
            that is being matched.
        :param min_distance:
            The maximum admissible distance from the epipolar line a candidate point
            should have to not be filtered out.
        """
        super().__init__(*args, **kwargs)
        self._fundamental_matrix = fundamental_matrix
        self._max_distance = max_distance

    def __call__(self, keypoint: Point, candidate_points: List[Point]) -> Generator[Point, None, None]:
        epipolar_line = get_epipolar_line(
            self._fundamental_matrix,
            self._matched_point_transfomer(keypoint)
        )

        return (
            candidate_point
            for candidate_point
            in candidate_points
            if point_to_line_distance(
                self._candidate_points_transformer(candidate_point),
                epipolar_line
            ) < self._max_distance
        )



from mmir.feature_matching import Correlations
from mmir.math import get_epipolar_line, line_intersection


class EpipolarTransfer:
    def __init__(self, left_to_destination, right_to_destination, transfer_right_matches: bool = True):
        self._left_to_destination = left_to_destination
        self._right_to_destionation = right_to_destination
        self._transfer_right_matches = transfer_right_matches

        self.destination_image = None

    def __call__(self, feature_matcher_output: Correlations) -> Correlations:
        flir_points = []

        for match in feature_matcher_output.matches:
            left_point = match[0]
            right_point = match[1]

            epi_line_from_left = get_epipolar_line(self._left_to_destination, left_point)
            epi_line_from_right = get_epipolar_line(self._right_to_destionation, right_point)

            point_in_flir = line_intersection(epi_line_from_left, epi_line_from_right)

            flir_points.append(point_in_flir)

        final_matches = []
        for i in range(0, len(flir_points)):
            final_matches.append((
                flir_points[i],
                feature_matcher_output.matches[i][1 if self._transfer_right_matches else 0],
                1
            ))

        feature_matcher_output.matches = final_matches
        feature_matcher_output.left = self.destination_image
        return feature_matcher_output



from .descriptor_filter import DescriptorsFilter
from .epipolar_distance_filter import EpipolarDistanceFilter
from .x_position_filter import XPositionFilter
from .y_position_filter import YPositionFilter



from typing import Any, Tuple

import numpy as np
from dataclasses import dataclass

from mmir import Matrix, Vector


@dataclass
class Camera:
    # The camera's intrinsic matrix.
    A: Any

    # The camera's translation vector.
    T: Any

    # The camera's rotation matrix.
    R: Any

    @staticmethod
    def from_files(intrinsic_parameters_file_path: str, extrinsic_parameters_file_path: str):
        return Camera(
            read_intrinsic_matrix(intrinsic_parameters_file_path),
            *read_extrinsic_parameters(extrinsic_parameters_file_path)
        )


def read_intrinsic_matrix(path: str) -> Matrix:
    """
    Reads the intrinsic parameters of a camera from a file, assuming the file's format is

    fx fy
    u v

    where fx and fy is the focal length in pixels,
    while (u, v) is the principal point of the camera.

    And returns them in matrix form.
    """
    with open(path) as f:
        lines = f.readlines()

        [fx, fy] = lines[0].split()
        [u, v] = lines[1].split()

        intrinsic_matrix = np.array([
            [fx, 0., u],
            [0., fy, v],
            [0., 0., 1.]
        ], dtype=float)

        return intrinsic_matrix


def read_extrinsic_parameters(path: str) -> Tuple[Vector, Matrix]:
    """
    Reads the extrinsic parameters of a camera from a file, assuming the file's format is

    tx ty tz
    r11 r12 r13
    r21 r22 r23
    r31 r32 r33

    where T = [tx ty tz].T is the translation vector of the camera

              [ r11 r12 r13 ]
    while R = [ r21 r22 r23 ] is the rotation matrix of the camera
              [ r31 r32 r33 ]

    and returns the tuple (T, R).
    """
    with open(path) as f:
        lines = f.readlines()

        tc = np.array(lines[0].split(), dtype=float)
        rc = np.array([line.split() for line in lines[1:4]], dtype=float)

        return tc, rc


def compute_fundamental_matrix(left_camera: Camera, right_camera: Camera) -> Matrix:
    """
    Computes the fundamental matrix relating two cameras
    """
    r_lr = right_camera.R.T @ left_camera.R
    [tx, ty, tz] = left_camera.R @ (right_camera.T - left_camera.T)

    s = np.array([
        [0, -tz, ty],
        [tz, 0, -tx],
        [-ty, tx, 0]
    ])

    e = r_lr @ s
    return np.linalg.inv(right_camera.A).T @ e @ np.linalg.inv(left_camera.A)



import abc
from typing import Callable, List, Generator

from mmir import Point
from mmir.math.math import identity


class DescriptorsFilter(abc.ABC):
    """
    Defines the interface a descriptor filter should implement.

    A descriptor filter is used to reduce the search space of the search
    for a match of a certain point.

    For example, one good example of a descriptor filter is one that for a given keypoint,
    it filters out all the candidates which are further away than a certain distance from
    the corresponding epipolar line of this point.
    """

    def __init__(
        self,
        matched_point_transformer: Callable[[Point], Point] = identity,
        candidate_points_transformer: Callable[[Point], Point] = identity
    ) -> None:
        """
        Constructs a DescriptorFilter.

        If one wants to transform the coordinates of the keypoint for which the descriptors are
        filtered, or transform the coordinates of the candidates themselves, before applying
        the filters, one can provide callables for these tasks through the following parameter:

        For example, if one of the image is scaled, and your filter works with epipolar lines, you
        might want to transform the points to their original coordinates before doing any
        computations.

        :param matched_point_transformer:
            Function to be called on the point for which candidates are filtered out,
            before applying the filter.
        :param candidate_points_transformer:
            Function to be called on the coordinates of every candidate point, before the
            filter is applied.
        """
        self._matched_point_transfomer = matched_point_transformer
        self._candidate_points_transformer = candidate_points_transformer

    @abc.abstractmethod
    def __call__(self, keypoint: Point, candidate_points: List[Point]) -> Generator[Point, None, None]:
        pass



import cv2
import numpy as np

from mmir import Image
from .disparity_map_calculator import DisparityMapCalculator, DisparityMap


class SGBMDisparityMapCalculator(DisparityMapCalculator):
    def __call__(self, left: Image, right: Image, **kwargs) -> DisparityMap:
        min_disp = kwargs.get('min_disp', 0)
        num_disp = kwargs.get('num_disp', 64 - min_disp)
        block_size = kwargs.get('block_size', 5)
        p1 = kwargs.get('p1', 16 * 3 ** 2)
        p2 = kwargs.get('p2', 64 * 3 ** 2)

        left_matcher = cv2.StereoSGBM_create(
            minDisparity=min_disp,
            numDisparities=num_disp,
            blockSize=block_size,
            P1=p1,
            P2=p2,
            mode=cv2.STEREO_SGBM_MODE_SGBM_3WAY,
        )

        right_matcher = cv2.ximgproc.createRightMatcher(left_matcher)
        wls_filter = cv2.ximgproc.createDisparityWLSFilter(left_matcher)

        left_disp = left_matcher.compute(left, right)
        right_disp = right_matcher.compute(right, left)

        disparity_mpap = wls_filter.filter(left_disp, left, None, right_disp).astype(np.float32) / 16.0
        return DisparityMap(left, right, disparity_mpap)

    @staticmethod
    def visualize(output: DisparityMap):
        cv2.imshow("disparity_map", output.disparity_map / 64.0)
        return output



import abc

from typing import List, Optional

from dataclasses import dataclass

from mmir.feature_detection import InterestPoints
from mmir.feature_detection.filters import DescriptorsFilter
from mmir import Image, Match


@dataclass
class Correlations:
    left: Image
    right: Image
    matches: List[Match]


class FeatureMatcher(abc.ABC):
    """
    Common interface to be used for the matching step of the registration pipeline.

    It defines the input and the output of this pipeline's step.

    It also defines a static method for basic visualization of the matching.
    """
    def __init__(self, filters: Optional[List[DescriptorsFilter]]):
        """
        :param filters:
            When figuring out the matching candidates of one point,
            optional filters will be applied to reduce the search space.

            For example, for a point on the left image, one might only
            consider the features from the right image close to the epipolar line
            corresponding to this point.
        """
        self._filters = filters

    @abc.abstractmethod
    def __call__(self, features: InterestPoints) -> Correlations:
        pass



import multiprocessing
from typing import Optional, List, Tuple

import cv2
import numpy as np

from joblib import Parallel, delayed
from mmir import Point, Image, Match
from mmir.feature_detection import InterestPoints
from mmir.feature_detection.filters import DescriptorsFilter
from mmir.feature_matching import FeatureMatcher, Correlations
from mmir.math import window_in_bounds, roi, get_epipolar_line, identity


def bresenham(x0, y0, x1, y1, transformer=identity):
    """Yield integer coordinates on the line from (x0, y0) to (x1, y1).

    Input coordinates should be integers.

    The result will contain both the start and the end point.
    """
    dx = x1 - x0
    dy = y1 - y0

    xsign = 1 if dx > 0 else -1
    ysign = 1 if dy > 0 else -1

    dx = abs(dx)
    dy = abs(dy)

    if dx > dy:
        xx, xy, yx, yy = xsign, 0, 0, ysign
    else:
        dx, dy = dy, dx
        xx, xy, yx, yy = 0, ysign, xsign, 0

    D = 2 * dy - dx
    y = 0

    for x in range(dx + 1):
        yield transformer((x0 + x * xx + y * yx, y0 + x * xy + y * yy))
        if D >= 0:
            y += 1
            D -= 2 * dx
        D += 2 * dy


class MIFeatureMatcher(FeatureMatcher):
    """
    Uses HOG (Histogram of Oriented Gradients) to find matches between
    two sets of keypoints of two images.
    """

    def __init__(
            self,
            window_size: int = 40,
            filters: Optional[List[DescriptorsFilter]] = None,
            fundamental_matrix=None,
            matched_point_point_transformer=identity,
            best_match_threshold=None,
            candidate_point_transformer=None,
    ) -> None:
        super().__init__(filters=filters)

        self._window_size = window_size
        self._filters = filters if filters else []
        self._fundamental_matrix = fundamental_matrix
        self._matched_point_transformer = matched_point_point_transformer
        self._best_match_threshold = best_match_threshold
        self._candidate_point_transformer = candidate_point_transformer

    def mi(self, left, right) -> float:
        hist_2d, _, _ = np.histogram2d(
            left.ravel(),
            right.ravel(),
            bins=48
        )
        pxy = hist_2d / float(np.sum(hist_2d))
        px = np.sum(pxy, axis=1)  # marginal for x over y
        py = np.sum(pxy, axis=0)  # marginal for y over x
        px_py = px[:, None] * py[None, :]  # Broadcast to multiply marginals
        nzs = pxy > 0  # Only non-zero pxy values contribute to the sum
        return float(1 - np.sum(pxy[nzs] * np.log(pxy[nzs] / px_py[nzs])))

    def _get_candidate_points(self, matched_point, right: Image):
        epipolar_line = get_epipolar_line(
            self._fundamental_matrix,
            self._matched_point_transformer(matched_point)
        )

        a_r = epipolar_line[0]
        b_r = epipolar_line[1]
        c_r = epipolar_line[2]

        p1x = int(0.0)
        p1y = int(-c_r / b_r)

        p2x = int(right.shape[1])
        p2y = int((-c_r - p2x * a_r) / b_r)

        p3x = int(-c_r / a_r)
        p3y = int(0)

        p4x = int((-b_r * right.shape[0] - c_r) / a_r)
        p4y = int(right.shape[0])

        p1x = max(p1x, p3x)
        p1y = max(p1y, p3y)

        p2x = min(p2x, p4x)
        p2y = min(p2y, p4y)

        return bresenham(p1x, p1y, p2x, p2y)

    def find_match(
            self,
            matched_point: Point,
            left: Image,
            right: Image
    ) -> Optional[Tuple[Point, float]]:
        if not window_in_bounds(center=matched_point, window_size=self._window_size, shape=left.shape):
            return None

        keypoint_subregion = roi(
            center=matched_point,
            width=self._window_size,
            height=self._window_size,
            image=left
        )

        candidate_points = self._get_candidate_points(matched_point, right)

        candidate_points = (
            descriptor for descriptor in candidate_points if window_in_bounds(
                center=descriptor,
                window_size=self._window_size,
                shape=left.shape
            )
        )

        for filter in self._filters:
            candidate_points = filter(matched_point, candidate_points)

        best_match = None
        for candidate_point in candidate_points:
            candidate_subregion = roi(
                center=candidate_point,
                width=self._window_size,
                height=self._window_size,
                image=right,
            )
            distance = self.mi(keypoint_subregion, candidate_subregion)
            if best_match is None or distance < best_match[1]:
                best_match = (candidate_point, distance)

        if not best_match:
            return None

        if self._best_match_threshold is None:
            return best_match

        if best_match[1] > self._best_match_threshold:
            return None

        return best_match

    def find_match_for(self, matched_point: Point, detector_output: InterestPoints) -> Optional[Match]:
        match = self.find_match(
            matched_point,
            detector_output.left,
            detector_output.right
        )

        if not match:
            return None

        return matched_point, match[0], match[1]

    @staticmethod
    def visualize(output: Correlations) -> Image:
        left = cv2.cvtColor(np.copy(output.left), cv2.COLOR_GRAY2BGR)
        right = cv2.cvtColor(np.copy(output.right), cv2.COLOR_GRAY2BGR)

        right = cv2.resize(src=right, dsize=(left.shape[1], left.shape[0]))
        img = np.concatenate((left, right), axis=1)
        for match in output.matches:
            pt1 = (int(match[0][0]), int(match[0][1]))
            pt2 = (int(match[1][0] + output.left.shape[1]), int(match[1][1]))

            cv2.line(img=img, pt1=pt1, pt2=pt2, color=(0, 255, 0))

        cv2.imshow("img", img)
        return output

    def __call__(self, detector_output: InterestPoints) -> Correlations:
        matches = Parallel(n_jobs=8)(
            delayed(self.find_match_for)(keypoint, detector_output) for keypoint in detector_output.keypoints_left
        )

        matches = filter(lambda x: x is not None, matches)
        matches = sorted(matches, key=lambda x: x[2], reverse=True)
        return Correlations(detector_output.left, detector_output.right, matches)



from typing import List, Optional

import cv2
import numpy as np

from mmir import Point, Image
from .feature_detector import FeatureDetector, InterestPoints


class GFTTFeatureDetector(FeatureDetector):
    """
    Feature detection using openCV's goodFeaturesToTrack function.
    """
    def __init__(self, subregions: int=1, points_per_region=30, skip_right=False, *args, **kwargs):
        """
        :param subregions:
            If you want to apply GFTT on subregions instead of the whole image,
            use this parameter too specify in how many subregions should the image
            be split in.
        """
        assert subregions > 0
        super().__init__(*args, **kwargs)

        self._points_per_region = points_per_region
        self._subregions = subregions
        self._skip_right = skip_right


    @staticmethod
    def visualize(output: InterestPoints) -> Image:
        left = np.copy(output.left)
        right = np.copy(output.right)

        left = cv2.cvtColor(left, cv2.COLOR_GRAY2BGR)
        right = cv2.cvtColor(right, cv2.COLOR_GRAY2BGR)

        for keypoint in output.keypoints_left:
            cv2.circle(left, (int(keypoint[0]), int(keypoint[1])), 1, (0, 0, 255))

        for keypoint in output.keypoints_right:
            cv2.circle(right, (int(keypoint[0]), int(keypoint[1])), 1, (0, 0, 255))

        cv2.imshow("featureDetector", np.concatenate((left, right), axis=1))
        return output

    def _filter_masked_interest_points(self, interest_points: List[Point], mask: Optional[Image]=None) -> List[Point]:
        """
        Excludes the interest points which are masked out, if a mask is specified.
        """
        if mask is None:
            return interest_points

        return [feature for feature in interest_points if mask[(feature[1], feature[0])]]

    def _interest_points_from_subregion(
        self,
        subregion: Image,
        x_offset: int,
        y_offset: int,
        mask: Optional[Image]=None
    ):
        """
        Applies GFTT on a subregion of the original image, and returns the interest points
        of this subregion.

        The offset parameters are used to map the coordinates of the interest points from
        the coordinate system of the subregion to the coordinate system of the entire image.

        :param subregion:
            The subregion on which GFTT should be applied.
        :param y_offset:
            The y coordinate of the region.
        :param x_offset:
            The x coordinate of the region.
        :return:
        """
        features = cv2.goodFeaturesToTrack(subregion, self._points_per_region, 0.1, 0)
        features = [feature[0] for feature in features] if features is not None else []
        features = [(int(feature[0]) + x_offset, int(feature[1]) + y_offset) for feature in features]
        return self._filter_masked_interest_points(features, mask)

    def _find_interest_points_in_subregions(
        self,
        left: Image,
        right: Image
    ) -> InterestPoints:
        """
        Finds the interest points of every subregion and unifies the results.
        """
        cols = left.shape[1]
        rows = left.shape[0]

        left_corners = []
        right_corners = []

        y_step = rows // self._subregions
        x_step = cols // self._subregions

        for y_offset in range(0, rows, y_step):
            for x_offset in range(0, cols, x_step):
                left_subregion = left[y_offset:y_offset + y_step, x_offset:x_offset + x_step]
                left_corners.extend(
                    self._interest_points_from_subregion(left_subregion, x_offset, y_offset, self._left_mask)
                )

                if not self._skip_right:
                    right_subregion = right[y_offset:y_offset + y_step, x_offset:x_offset + x_step]
                    right_corners.extend(
                        self._interest_points_from_subregion(right_subregion, x_offset, y_offset, self._right_mask)
                    )

        output = InterestPoints(left, right, left_corners, right_corners)

        return output

    def __call__(self, left: Image, right: Image) -> InterestPoints:
        if self._subregions != 1:
            return self._find_interest_points_in_subregions(left, right)

        left_keypoints = cv2.goodFeaturesToTrack(left, 300, 0.0000001, 0, mask=self._left_mask)
        left_keypoints = [kp[0] for kp in left_keypoints]

        right_keypoints = []
        if not self._skip_right:
            right_keypoints = cv2.goodFeaturesToTrack(right, 500, 0.000000001, 0, mask=self._right_mask)
            right_keypoints = [kp[0] for kp in right_keypoints]

        output = InterestPoints(left, right, left_keypoints, right_keypoints)

        return output



import setuptools

with open("README.md") as fh:
    long_description = fh.read()


setuptools.setup(
    name="mmir",
    version="0.1.2",
    author="Alexandru Kis",
    author_email="alexandrukis1618033@gmail.com",
    description="A package for helping you register VIS and LWIR images",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/_alexandrukis/pyreg/src/master/",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ),
    install_requires=[
        'bresenham==0.2',
        'dataclasses==0.6',
        'funcy==1.10.2',
        'joblib==0.12.0',
        'multipledispatch==0.5.0',
        'numpy==1.14.5',
        'opencv-contrib-python==3.4.1.15',
        'scikit-image==0.14.0',
        'scipy==1.1.0',
    ]
)



import abc
from typing import Any

from dataclasses import dataclass

from mmir import Image
from mmir.feature_matching import Correlations


@dataclass
class EstimatedMappingModel:
    left: Image
    right: Image
    model: Any


class MappingEstimator(abc.ABC):
    @abc.abstractmethod
    def __call__(self, correlation: Correlations) -> EstimatedMappingModel:
        pass



import os
import cv2

from mmir.camera.camera import Camera, compute_fundamental_matrix
from mmir.mapping import MappingVisualizer
from mmir.pipeline import Pipeline
from mmir.pipeline.interceptor import Interceptor
from mmir.stereo.disparity_transfer import MappingCalculator, OneToOneMapping
from mmir.stereo.sgbm_disparity_map_calculator import SGBMDisparityMapCalculator

RSC_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../rsc')


def yield_all_photos():
    left_photos = [os.path.join(RSC_PATH, f) for f in os.listdir(RSC_PATH) if f.endswith("-1.bmp")]
    left_photos.sort()

    right_photos = [os.path.join(RSC_PATH, f) for f in os.listdir(RSC_PATH) if f.endswith("-2.bmp")]
    right_photos.sort()

    flir_photos = [os.path.join(RSC_PATH, f) for f in os.listdir(RSC_PATH) if f.endswith("-7.bmp")]
    flir_photos.sort()

    for i in range(0, len(left_photos)):
        left = cv2.imread(left_photos[i], cv2.IMREAD_GRAYSCALE)
        right = cv2.imread(right_photos[i], cv2.IMREAD_GRAYSCALE)
        flir = cv2.imread(flir_photos[i], cv2.IMREAD_GRAYSCALE)

        yield left, right, flir


def main():
    flir_camera = Camera.from_files(
        os.path.join(RSC_PATH, 'left_int_F410.txt'),
        os.path.join(RSC_PATH, 'left_ext_F410.txt'),
    )

    left_camera = Camera.from_files(
        os.path.join(RSC_PATH, 'left_int.txt'),
        os.path.join(RSC_PATH, 'left_ext.txt'),
    )

    right_camera = Camera.from_files(
        os.path.join(RSC_PATH, 'right_int.txt'),
        os.path.join(RSC_PATH, 'right_ext.txt'),
    )

    left_to_flir = compute_fundamental_matrix(left_camera, flir_camera)
    right_to_flir = compute_fundamental_matrix(right_camera, flir_camera)

    interceptor = Interceptor()
    pipeline = Pipeline([
        SGBMDisparityMapCalculator(),
        SGBMDisparityMapCalculator.visualize,
        MappingCalculator(left_to_flir, right_to_flir),
        interceptor,
        MappingVisualizer(fuse=True),
    ])

    for (left, right, flir) in yield_all_photos():
        interceptor.intercept = lambda output: OneToOneMapping(output.left, flir, output.mapping)
        pipeline.run(left, right)


if __name__ == '__main__':
    main()



from typing import Callable, List, Generator, Optional, Union

from mmir import Point
from .descriptor_filter import DescriptorsFilter


class YPositionFilter(DescriptorsFilter):
    """
    Filters out all the descriptors which have y coordinate
    outside the range (lower_bound(matched_point.y), upper_bound(matched_point.y)).
    """

    def __init__(
            self,
            lower_bound: Optional[Union[Callable[[int], int], int]],
            upper_bound: Optional[Union[Callable[[int], int], int]],
            *args,
            **kwargs
    ) -> None:
        """
        :param lower_bound:
            Callable to compute the lower bound of the interval of admissible y values,
            relative to the currently matched point.
        :param upper_bound:
            Callable to compute the upper bound of the interval of admissible y values,
            relative to the currently matched point.
        """
        super().__init__(*args, **kwargs)
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    def __call__(self, matched_point: Point, candidate_points: List[Point]) -> Generator[Point, None, None]:
        y = matched_point[1]
        lower_bound = None
        if self._lower_bound:
            lower_bound = self._lower_bound(y) if callable(self._lower_bound) else y + self._lower_bound

        upper_bound = None
        if self._upper_bound:
            upper_bound = self._upper_bound(y) if callable(self._upper_bound) else y + self._upper_bound

        return (
            candidate_point
            for candidate_point in candidate_points
            if (lower_bound is None or lower_bound < candidate_point[1]) and (
                upper_bound is None or candidate_point[1] < upper_bound)
        )
